/*****************************************************************
 ** Copyright 1999 Clark Cooper
 ** All rights reserved.
 ** This is free software; you may modify it or redistribute it
 ** under the terms of the perl Artistic license.
 */

#ifndef _HASH_H
#define _HASH_H

#include <stdio.h>

/*================================================================
** Public macro definitions
*/

#define HASH_MAXKEYS	8

#define HASH_CHECK_CNT(x) if((x)>=HASH_MAXKEYS)\
	{fprintf(stderr,"Specified too many keys for hash table.\n");exit(-1);}

#define HASH_BEGIN(type) {type dmy; int _c=0; KInfo _k[HASH_MAXKEYS]
#define HASH_KEY(comp) HASH_CHECK_CNT(_c) \
		       _k[_c].siz=sizeof(dmy.comp);\
		       _k[_c++].off=(char*)&dmy.comp-(char*)&dmy
#define HASH_KEY_STR(comp) HASH_CHECK_CNT(_c) \
			   _k[_c].siz=-sizeof(dmy.comp);\
			   _k[_c++].off=(char*)&dmy.comp-(char*)&dmy
#define HASH_KEY_PSTR(comp) HASH_CHECK_CNT(_c) \
			   _k[_c].siz=0;\
                           _k[_c++].off=(char*)&dmy.comp-(char*)&dmy
#define HASH_END(var, isize) var=newhash(isize,_c,_k);}

/*================================================================
** Public type definitions
*/

typedef struct _ki	KInfo;
typedef	struct _htable	HashTable;
typedef void		*HashObj;
typedef const void	*ConstHashObj;
typedef void		*HashIterator;

/*================================================================
** Public structure declarations
*/

struct _ki
{
  short	off;
  short	siz;
};

/*================================================================
** Public function declarations
*/

/*--------------------------------------------------
  Return a hash table with room for at least initial_size
  elements. If kcnt <= 0, kin may be NULL and the key is either a
  null-terminated string or a structure that begins with a null-terminated
  string (e.g. struct {char x[16]; int foo} could have kcnt == 0 if all
  instances of x were null terminated, struct {char *y; int foo;} couldn't
  have kcnt == 0).

  Otherwise kcnt is the number of keys specified in kin. Except in the
  simple case where kcnt == 0, newhash shouldn't be called directly.
  Instead callers should use the macros HASH_BEGIN, HASH_KEY, HASH_KEY_STR,
  and HASH_END defined in hash.h.

  Returns NULL if there's not enough memory to allocate some part of
  the hashtable or if kcnt > HASH_MAXKEYS.
  */

extern HashTable *
newhash(
	int		initial_size,	/* # Elements to hold */
	int		kcnt,		/* Number of keys */
	KInfo		*kinf		/* Key info */
	);


/*--------------------------------------------------
  Return the HashObj matching the key. If no entry matches,
  return NULL.
  */

extern HashObj
hashfind(HashTable *ht, ConstHashObj key);

/*--------------------------------------------------
  If a HashObj is found matching the key, return that entry.
  Otherwise insert the entry provided as a key and return it.
  */

extern HashObj
hashinsert(HashTable *ht, ConstHashObj key);

/*--------------------------------------------------
  If an entry is found matching the key, delete that entry from
  the table and return it. Otherwise return NULL.
  */

extern HashObj
hashdelete(HashTable *ht, ConstHashObj key);

/*--------------------------------------------------
  Delete all entries from hash table.
  */

extern void
hashclear(HashTable *ht);

/*--------------------------------------------------
  Free hash table and internal memory assocated with it.
  This does not free the memory of HashObj entries, which
  are the responsibility of the caller. If the hash table
  contains the only references to the object, you may free
  the objects (before calling hashclear or hashfree) by
  iterating over them with the hashnext function described below.
  */

extern void
hashfree(HashTable *ht);

/*--------------------------------------------------
  Return the number of entries in the table.
  */

extern int
hashcount(HashTable *ht);

/*--------------------------------------------------
  Place a pointer to a newly allocated vector of hash objects
  containing all the entries in the table in the variable pointed
  at by vec. Return value is the number of objects in this vector.
  The caller is responsible for freeing the vector.

  If there isn't enough memory to allocate the vector, then a -1
  is returned and vec is set to NULL. If the table is empty, then
  the variable pointed at by vec isn't changed.
  */

extern int
hashvector(HashTable *ht, HashObj **vec);

/*--------------------------------------------------
  This function is used for iterating over all the
  entries in the hash table. To start off, set the
  value pointed to by ITER to 0. The function will
  return the next hash object from the table. It returns
  NULL when there are no more entries (and sets the HashIterator
  to which ITER points to 0). Note that hash tables
  do not order their objects. Note that ITER should not be
  NULL. (However hashnext will return NULL if you
  do pass a NULL value for ITER).

  Here's an example of how to use:

  HashIterator	step = 0;
  HashObj	obj;

  while ((obj = hashnext(htable, &step)) != NULL)
  {
    ... 
  }

  */

extern HashObj
hashnext(HashTable *ht, HashIterator *iter);

#endif  /* ndef _HASH_H */
